from setuptools import setup

setup(
    name="tapystry",
    packages=["tapystry"],
    version="0.3.18",
    author="Jeffrey Wu",
    install_requires=[
    ],
    url="https://github.com/wuthefwasthat/tapystry",
    description="A redux-saga inspired effects library",
    python_requires='>=3.7',
)
